/* -*- c++ -*-
   god.cc

   General Operations Director

   perform operations requiring omnipotence in the simulation

   NOTE: Tcl node indexs are 0 based, NS C++ node IP addresses (and the
   node->index() are 1 based.

   $Id: god.cc,v 1.1.1.1 2000/08/28 18:40:06 jinyang Exp $
   */

#include <object.h>
#include <packet.h>
#include <ip.h>

#include <god.h>

#define RANGE 250.0
#define BIGGEST_MASK 32
#define MAX_Q_LEN 100

#define WHITE 0
#define GREY 1
#define BLACK 2

God* God::instance_;

static class GodClass : public TclClass {
public:
        GodClass() : TclClass("God") {}
        TclObject* create(int, const char*const*) {
                return (new God);
        }
} class_God;


God::God()
{
        min_hops = 0;
        num_nodes = 0;

}


int
God::hops(int i, int j)
{
        return min_hops[i * num_nodes + j];
}


void
God::stampPacket(Packet *p)
{
        hdr_cmn *ch = HDR_CMN(p);
        struct hdr_ip *ih = HDR_IP(p);
        nsaddr_t src = ih->src();
        nsaddr_t dst = ih->dst();

        assert(min_hops);

        if (!DATA_PACKET(ch->ptype())) return;

        if (dst > num_nodes || src > num_nodes) return; // broadcast pkt
   
        ch->opt_num_forwards() = min_hops[src * num_nodes + dst];
}


void 
God::recv(Packet *, Handler *)
{
    abort();
}

/* added by Jinyang */
bool
God::isAwake(nsaddr_t index)
{
   double now = Scheduler::instance().clock();
   if (now > wakelist[index]) {
       return false;
    }else {
       return true;
    }
}

/* added by jinyang 
   testing whether two nodes are within reasonable communication distance from each other
   if so, the link error for the routing protocol might indicate something other than
   the next hop moving away, e.g. congestion */
bool
God::isNeighbor(nsaddr_t src, nsaddr_t dst)
{
    double src_x,src_y,src_z;
    double dst_x,dst_y,dst_z;
    
    (nodelist[src])->getLoc(&src_x,&src_y,&src_z); 
    (nodelist[dst])->getLoc(&dst_x,&dst_y,&dst_z); 

    double dist = (src_x - dst_x) * (src_x - dst_x) + (src_y - dst_y) * (src_y - dst_y);
    dist = sqrt(dist);

    if (dist < (RANGE - 10.0)) {
	return true;
    }else{
	return false;
    }
}


#ifdef IDEAL_ROUTE
/* god should know the route to anywhere -Jinyang*/
bool
God::getRouteToLoc(nsaddr_t fromnode, location dstloc, nsaddr_t *route, int maxnodes)
{
    nsaddr_t u,node;
    Quadtree *mytree;
    Quadtree *tree;
    qtree_node *ptr;
    int which_nb,ns,ew;
    double x,y,z,my_x,my_y,my_z,min_x,max_x,min_y,max_y;
    int q_head,q_len,q_tail;
    nsaddr_t dst = 0;
    int stop,power,the_loc;
    int color[num_nodes];
    nsaddr_t predecessor[num_nodes];
    nsaddr_t queue[MAX_Q_LEN];

    //God does not balance the load :-) maybe he should?
    //it just computes the shortest path
    
    //breadth-first search for the dst loc
    //set adjacency matrix
    
    //i suspect this would be slow too

    //convert the dstloc to min_x, max_x,min_y,max_y
    stop = BIGGEST_MASK;
    power = 1;
    while (stop >=2) {
	y = y + power * (the_loc & 0x0001);
	the_loc = the_loc >> 1;
	x = x + power * (the_loc & 0x0001);
	the_loc = the_loc >> 1;
	stop = stop - 2;
        power = power * 2;
    }

    min_x = x * RANGE;
    min_y = y * RANGE;
    //reuse the power variable
    power = (BIGGEST_MASK - dstloc.mask)/2;
    max_x = min_x + RANGE*(pow(2,power));
    max_y = min_y + RANGE*(pow(2,power));

    for (int i=0;i<num_nodes;i++) {
	if (i != fromnode) {
	    color[i] = WHITE;
	    predecessor[i]=0;
	}
    }
    color[fromnode] = GREY;

    //enqueue the source
    q_head = 0;
    q_tail = 0;
    q_len = 1;
    queue[q_head] = fromnode;

    
    while (q_len) {
	u = queue[q_head];
	nodelist[u]->getLoc(&my_x,&my_y,&my_z);
	mytree = nodelist[u]->mytree;

	//get all the neighbors of u

	// get neighbors in the same grid
	ptr = (mytree->head).lh_first;
	for (;ptr;ptr=ptr->node_link.le_next) {

	    node = (nsaddr_t)((MobileNode *)(ptr->node))->index_;

	    if (node == u || (color[node] != WHITE)) {
		continue;
	    }

	    (nodelist[node])->getLoc(&x,&y,&z); 
	    if ((sqrt(x-my_x)*(x-my_x)+(y-my_y)*(y-my_y)) < RANGE) {

		predecessor[node] = u;
		color[node] = GREY;

		if ((x>min_x) && (x<max_x) && (y>min_y) && (y<max_y)) {
		    dst = node;
		    goto NEXT;
		}

		//enqueue my neighbor
		q_tail = (q_tail + 1) % MAX_Q_LEN;
		q_len++;
		assert(q_len <= MAX_Q_LEN);
		queue[q_tail] = node;
	    }
	}

	//get neighbors in adjacent grid
	for (ns=0;ns<3;ns++) {
	    for (ew=0;ew<3;ew++) {

		which_nb = ns*4+ew;
		if (which_nb == 0) continue;

		tree = mytree->getNeighbor(which_nb);
		if (tree==NULL) continue;

		ptr = (tree->head).lh_first;
		for (;ptr;ptr=ptr->node_link.le_next) {

		   node = ((MobileNode *)(ptr->node))->index_;

		   if (color[node] != WHITE) continue;

		   nodelist[node]->getLoc(&x,&y,&z); 
		   if ((sqrt(x-my_x)*(x-my_x)+(y-my_y)*(y-my_y)) < RANGE) {
		       color[node] = GREY;
		       predecessor[node] = u;

		       if ((x>min_x) && (x<max_x) && (y>min_y) && (y<max_y)) {
			   dst = node;
			   goto NEXT;
		       }
		       //enqueue my neighbor
		       q_tail = (q_tail + 1) % MAX_Q_LEN;
		       q_len++;
		       assert(q_len <= MAX_Q_LEN);
		       queue[q_tail] = node;
		    }

		}
	    }
	}

	//dequeue u
	q_head = (q_head+1) % MAX_Q_LEN;
	q_len--;
	assert(q_len>=0);
	color[u] = BLACK;
    }

    return false;

NEXT:
    //fill the route
    assert(dst != 0);
    nsaddr_t *tmp = new nsaddr_t[maxnodes];
    int i,j;
    i = 0;
    while (dst!= fromnode) {
	tmp[i++] = dst;
	dst = predecessor[dst];
	assert(dst != 0);
	assert(i<maxnodes);
    }
    tmp[i]= fromnode;

    assert(i<=maxnodes);
    //inverse it!
    j = 0;
    while (i>=0) {
	route[j++] = tmp[i--];
    }
    delete tmp;
    return true;
}
#endif


int 
God::command(int argc, const char* const* argv)
{
        if (argc == 2) {
                if(strcmp(argv[1], "dump") == 0) {
                        int i, j;

                        for(i = 1; i < num_nodes; i++) {
                                fprintf(stdout, "%2d) ", i);
                                for(j = 1; j < num_nodes; j++)
                                        fprintf(stdout, "%2d ",
                                                min_hops[i * num_nodes + j]);
                                fprintf(stdout, "\n");
                        }
                        return TCL_OK;
                }
        }
        else if(argc == 3) {
                if (strcasecmp(argv[1], "num_nodes") == 0) {
                        assert(num_nodes == 0);

                        // allow for 0 based to 1 based conversion
                        num_nodes = atoi(argv[2]) + 1;

                        min_hops = new int[num_nodes * num_nodes];
                        bzero((char*) min_hops,
                              sizeof(int) * num_nodes * num_nodes);

                        instance_ = this;

			wakelist = new double[num_nodes];
			wakeup_list = new double[num_nodes];
			for (int i=0;i<num_nodes;i++) {
			    wakelist[i] = 0.0;
			    wakeup_list[i] = 0.0;
			}
			
			// allocate space to contain the reference to all the nodes
			//nodelist = 0;			
			nodelist = new MobileNode*[num_nodes];

#ifdef IDEAL_ROUTE
			assert(T!=NULL);
			double maxX;
			double maxY;
			double len = RANGE;
			int level = 0;

			maxX = T->upperX();
			maxY = T->upperY();

			while ((len < maxX) || (len < maxY)) {
			    level ++;
			    len *= 2;
			}
			root = new Quadtree(T->lowerX(),T->lowerY(),len,len,level,NULL,0);
#endif
                        return TCL_OK;

                }
#ifdef IDEAL_ROUTE
		else if(strcmp(argv[1], "topography") == 0) {
			T = (Topography*) TclObject::lookup(argv[2]);
			if(T == 0)
				return TCL_ERROR;
			return TCL_OK;
		}
#endif
		
        }
	else if (argc == 4) {
	        /* added by Jinyang to take care of wake-sleep scenarios
		*/
		if (strcasecmp(argv[1],"wakeup") == 0 ) {
		    int index = atoi(argv[2]);    
		    //update the wakelist accordingly
		    wakelist[index] = Scheduler::instance().clock()+ atof(argv[3]);

		    //when this guy has woken up
		    wakeup_list[index] = Scheduler::instance().clock();

		    return TCL_OK;
		}
	}
        else if(argc == 5) {
                if (strcasecmp(argv[1], "set-dist") == 0) {
                        int i = atoi(argv[2]);
                        int j = atoi(argv[3]);
                        int d = atoi(argv[4]);

                        assert(i > 0 && i < num_nodes);
                        assert(j > 0 && j < num_nodes);

                        min_hops[i * num_nodes + j] = d;
                        min_hops[j * num_nodes + i] = d;
                        return TCL_OK;
                }
        } 
        return Connector::command(argc, argv);
}
