#include "hgpsrouter.h"

// Returns a random number between 0 and max, implemented by dmaltz
static inline double
jitter (double max, int be_random_)
{
      return (be_random_ ? Random::uniform(max) : 0);
}

static bool
inSameGrid(location src, location dst)
{
     int min_mask;
     int mask;
     
     if (src.mask > dst.mask) {
	 min_mask = dst.mask;
     }else{
	 min_mask = src.mask;
     }
    
     mask = 0xffffffff << (BIGGEST_MASK - min_mask);
     return (((src.loc & mask) == (dst.loc & mask))? true:false);
}


//this function converts grid location to a geographical distance
static inline void
loctodistance(int &x, int &y, location loc)
{
    int stop =loc.mask;
    int shift = BIGGEST_MASK;
    unsigned int tmp_mask = 0xffffffff;
    int the_loc = loc.loc;

    x = 0;
    y = 0;

    while (stop>=2) {
	shift = shift - 2;
	x = x *2 + 2* (the_loc >> (shift+1));
	y = y *2 + 2* ((the_loc & 0x55555555)>>shift);
	stop -= 2;
	//clear the examined two bits
	tmp_mask = tmp_mask >> 2;
	the_loc = the_loc & tmp_mask;
    }
    if (shift == 0) {
	x = x +1;
	y = y +1;
    }else {
	x = x*((int)pow(2,(shift/2)));
	x = x+((int)pow(2,(shift/2)));
	y = y*((int)pow(2,(shift/2)));
	y = y+((int)pow(2,(shift/2)));
    }
#ifdef HGPS_DEBUG100
    assert((x>=1)&&(x<=7));
    assert((y>=1)&&(y<=7));
#endif
#ifdef HGPS_DEBUG150
    if (loc.mask == 32) {
	assert((x>=1)&&(x<=11));
	assert((y>=1)&&(y<=11));
    }
#endif

    
}

HGPSRouterPeriodicHandler::HGPSRouterPeriodicHandler(HGPSRouter *r)
{
    myrouter = r;
}

void
HGPSRouterPeriodicHandler::handle(Event *e)
{
    //send out one hop neighbor table periodically!
    myrouter->broadcastOneHopRT();
    //reschedule the next periodic event
    Scheduler::instance().schedule(this,e,BROADCAST_INTEVAL);
}

HGPSRouter::HGPSRouter(HGPSAgent *a)
{
    parent = a;
    //initializing my routing table
    for (int i=0;i<MAX_RT_ENTRIES;i++) {
	nt[i].lastupdate_time = -2*NEIGHBOR_TIMEOUT_INTEVAL;
	nt[i].next_hop = NULL;
    }
    for (int i=0;i<MAX_ADJ_GRID;i++) {
	rt_1[i].loc.loc = -1;
	rt_1[i].lastupdate_time = -2*NEIGHBOR_TIMEOUT_INTEVAL;
	rt_1[i].next_hop = NULL;
    }

    router_periodic_event = new Event();
    router_periodic_handler = new HGPSRouterPeriodicHandler(this);
}

void
HGPSRouter::setLocServer(LocServer *s)
{
    locserver = s;
}
void
HGPSRouter::start()
{
    //periodic broadcasting
    Scheduler::instance().schedule(router_periodic_handler,
	    router_periodic_event,jitter(HGPS_ROUTER_STARTUP_JITTER, 1));
}

void
HGPSRouter::sleep()
{
    Scheduler::instance().cancel(router_periodic_event);
}

void
HGPSRouter::broadcastOneHopRT()
{
    Scheduler & s = Scheduler::instance();
    double now = s.clock();
    double x, y;

    Packet *pkt = parent->allocpkt();
    hdr_hgps *hgpsh =  (hdr_hgps*)pkt->access(parent->off_hgps_);
    hdr_ip *iph =  (hdr_ip*)pkt->access(parent->off_ip_); 
    hdr_cmn *cmnh =  (hdr_cmn*)pkt->access(parent->off_cmn_);
    location myloc;
    
    int num = 0;

    //send out all my one hop neighbors from routing table,
    //be careful about out-dated entries
    hgpsh->init();
    hgpsh->type_ = ROUTE_UPDATE;
    myloc = locserver->getGridPos(x,y);
    nsaddr_t hop;
    for (int i=0;i<MAX_RT_ENTRIES;i++) {
	if (((now - nt[i].lastupdate_time) < NEIGHBOR_TIMEOUT_INTEVAL) && 
		(hop=getNextHopNode(nt[i])) && (hop == nt[i].dst)) {
    	    hgpsh->neighbor_ids_[num] = nt[i].dst;
    	    hgpsh->neighbor_locs_[num] = nt[i].loc;
    	    num++;
	}
    }
    assert(num <= MAX_NEIGHBOR);
    hgpsh->neighbor_num_ = num;
    
    //update my own location
    hgpsh->src_loc_ = myloc;
    hgpsh->src_x_ = x;
    hgpsh->src_y_ = y;
    hgpsh->src_speed_ = locserver->getSpeed();

    //update commond headers etc.
    cmnh->next_hop_ = MAC_BROADCAST;
    cmnh->addr_type_ = AF_INET;
    cmnh->ptype() = PT_HGPS;

    
    iph->dst() = IP_BROADCAST;
    iph->dport_ = RT_PORT;
    iph->src() = parent->myaddr_;

    //ask loc server to piggy back any fp information they want
    locserver->piggybackFP(pkt);

    //update the pkt size
    cmnh->size() = hgpsh->size();

    //send out the packet immediately, yeah!!!!!!!
    s.schedule(parent->ll,pkt,0);
}

void
HGPSRouter::recvNeighborUpdate(Packet *pkt)
{
    hdr_hgps *hgpsh =  (hdr_hgps*)pkt->access(parent->off_hgps_);
    hdr_ip *iph =  (hdr_ip*)pkt->access(parent->off_ip_); 

    //update my little routing tables :P
    addUpdateEntry(hgpsh->src_loc_,iph->src(),hgpsh->src_x_,hgpsh->src_y_,iph->src(),1, hgpsh->src_speed_);

    for (int i=0;i<hgpsh->neighbor_num_;i++) {
	addUpdateEntry(hgpsh->neighbor_locs_[i],iph->src(),hgpsh->src_x_,hgpsh->src_y_,hgpsh->neighbor_ids_[i],2, 0);
    }

    if (hgpsh->fp_num_>0) {
	locserver->recvFPUpdate(pkt);
    }
    Packet::free(pkt);
}

void
HGPSRouter::deleteEntry(nsaddr_t nextnode)
{

    nexthop_entry *node_pt;
    nexthop_entry *node_tmp_pt;
    double now = Scheduler::instance().clock();

    for (int i = 0;i<MAX_RT_ENTRIES;i++) {
	node_pt = nt[i].next_hop;
	while (node_pt) {
	    if (((now - node_pt->lastupdate_time) > NEIGHBOR_TIMEOUT_INTEVAL)||(node_pt->next_node == nextnode)) {
		//delete delete
		if (node_pt->prev) {
		    node_pt->prev->next = node_pt->next;
		}else{
		    nt[i].next_hop = node_pt->next;
		}

		if (node_pt->next) {
		    node_pt->next->prev = node_pt->prev;
	        }
		node_tmp_pt = node_pt;
		node_pt = node_pt->next;
		delete node_tmp_pt; 
	    }else{
		node_pt = node_pt->next;
	    }
	}
	if (!nt[i].next_hop) {
	    nt[i].dst = 0;
	    nt[i].lastupdate_time = -10*NEIGHBOR_TIMEOUT_INTEVAL;
	}
    }

    
    //delete location entry
    for (int i=0;i<MAX_ADJ_GRID;i++) {
	node_pt = rt_1[i].next_hop;
	while (node_pt) {
	    if (((now - node_pt->lastupdate_time) > NEIGHBOR_TIMEOUT_INTEVAL)||(node_pt->next_node == nextnode)) {
		//delete delete
		if (node_pt->prev) {
		    node_pt->prev->next = node_pt->next;
		}else{
		    rt_1[i].next_hop = node_pt->next;
		}

		if (node_pt->next) {
		    node_pt->next->prev = node_pt->prev;
	        }

		node_tmp_pt = node_pt;
		node_pt = node_pt->next;
		delete node_tmp_pt; 
	    }else{
		node_pt = node_pt->next;
	    }
	}
	if (!rt_1[i].next_hop) {
	    rt_1[i].dst = 0;
	    rt_1[i].lastupdate_time = -10*NEIGHBOR_TIMEOUT_INTEVAL;
	}
    }
}

void
HGPSRouter::deleteEntry(nsaddr_t fromnode, nsaddr_t tonode)
{

    nexthop_entry *node_pt;
    nexthop_entry *node_tmp_pt;
    double now = Scheduler::instance().clock();
    int j;

    if (fromnode == tonode) {
	return;
    }

    //delete delete
    for (int i = 0;i<MAX_RT_ENTRIES;i++) {
	j = (i + tonode) % MAX_RT_ENTRIES;
	if ((nt[j].dst == tonode)) {
	    node_pt = nt[j].next_hop;
	    while (node_pt) {
		if (((now - node_pt->lastupdate_time) > NEIGHBOR_TIMEOUT_INTEVAL)||(node_pt->next_node == fromnode)) {
		    //delete delete
		    if (node_pt->prev) {
			node_pt->prev->next = node_pt->next;
		    }else{
			nt[j].next_hop = node_pt->next;
		    }

		    if (node_pt->next) {
			node_pt->next->prev = node_pt->prev;
		    }
		    node_tmp_pt = node_pt;
		    node_pt = node_pt->next;
		    delete node_tmp_pt; 
		}else{
		    node_pt = node_pt->next;
		}
	    }
	    if (!nt[j].next_hop) {
		nt[j].dst = 0;
		nt[j].lastupdate_time = -3*NEIGHBOR_TIMEOUT_INTEVAL;
	    }
	    return;
	}
    }

}


void
HGPSRouter::addUpdateEntry(location loc,nsaddr_t next_hop,double x, double y,nsaddr_t dst, int distance, double speed)
{
    double now = Scheduler::instance().clock();
    int freeslot = -1;
    int i;
   
    assert(dst>0);

    nexthop_entry *nexthop_pt;
    nexthop_entry *new_pt;

    for (int j=0;j<MAX_RT_ENTRIES;j++) {
	i = (j + dst) % MAX_RT_ENTRIES;
    	if ((nt[i].dst== dst)) {
	    //update the entry
	    nt[i].lastupdate_time = now;
	    nt[i].loc = loc;
	    nexthop_pt = nt[i].next_hop;
	    if ((distance == 1) || (!nexthop_pt)) {
		//insert in the front of the list
		if (nexthop_pt && (nexthop_pt->next_node == next_hop)){
		    //the entry already exists, just update the timer
		    nexthop_pt->lastupdate_time = now;
		    nexthop_pt->x = x;
		    nexthop_pt->y = y;
		    nexthop_pt->speed = speed;
		}else {
		    new_pt = new nexthop_entry;
		    new_pt->dist = distance;
		    new_pt->next_node = next_hop;
		    new_pt->lastupdate_time = now;
		    new_pt->x = x;
		    new_pt->y = y;
		    new_pt->speed = speed;
		    new_pt->prev = NULL;
		    new_pt->next = nexthop_pt;
		    nt[i].next_hop = new_pt;
		    if (nexthop_pt) {
		       nexthop_pt->prev = new_pt;
		    }
		}
	    }else{
		while((nexthop_pt->next)&&(nexthop_pt->next_node != next_hop)) {
		    nexthop_pt = nexthop_pt->next;
		}
		if (nexthop_pt->next_node == next_hop) {
		    nexthop_pt->x = x;
		    nexthop_pt->y = y;
		    nexthop_pt->lastupdate_time = now;
		    nexthop_pt->speed = speed;
		}else{
		    assert(nexthop_pt->next == NULL);
		    //insert the node at the end of nexthop_pt, the last entry
		    new_pt = new nexthop_entry;
		    new_pt->dist = distance;
		    new_pt->next_node = next_hop;
		    new_pt->lastupdate_time = now;
		    new_pt->speed = speed;
		    new_pt->x = x;
		    new_pt->y = y;
		    nexthop_pt->next = new_pt;
		    new_pt->next = NULL;
		    new_pt->prev = nexthop_pt;
		}
	    }
	    goto NEXT;
	}else if ((freeslot == -1) && ((now - nt[i].lastupdate_time) > NEIGHBOR_TIMEOUT_INTEVAL)) {
	    nexthop_pt = nt[i].next_hop;
	    //delete the next_hop list
	    while(nexthop_pt) {
		nexthop_pt = nexthop_pt->next;
		if (nexthop_pt) {
		    delete nexthop_pt->prev;
		}
	    }
	    nt[i].next_hop = NULL;
	    freeslot = i;
    	}
    }

    assert(freeslot>=0);

    //add it to freeslot
    new_pt = new nexthop_entry;
    new_pt->lastupdate_time = now;
    new_pt->x = x;
    new_pt->y = y;
    new_pt->speed = speed;
    new_pt->next_node = next_hop;
    new_pt->dist = distance;
    
    new_pt->prev = NULL;
    new_pt->next = NULL;
    
    nt[freeslot].loc = loc;
    nt[freeslot].next_hop = new_pt;
    nt[freeslot].lastupdate_time = now;
    nt[freeslot].dst = dst;

NEXT:    freeslot = -1;
    //update adjacent location grid
    for (int i = 0;i<MAX_ADJ_GRID;i++) {

	if (rt_1[i].loc.loc == loc.loc) {
	    rt_1[i].lastupdate_time = now;
	    nexthop_pt = rt_1[i].next_hop;
	    if (!nexthop_pt) {
		new_pt = new nexthop_entry;
		new_pt->dist = distance;
		new_pt->next_node = next_hop;
		new_pt->lastupdate_time = now;
		new_pt->speed = speed;
		new_pt->x = x;
		new_pt->y = y;
		new_pt->prev = NULL;
		new_pt->next = NULL;
		rt_1[i].next_hop = new_pt;
	    }else{
		while (nexthop_pt->next && (nexthop_pt->next_node != next_hop) 
			&& (nexthop_pt->dist <= distance)) {
		    nexthop_pt = nexthop_pt->next;
		}
	
		if (nexthop_pt->next_node == next_hop) {
		    nexthop_pt->lastupdate_time = now;
		    nexthop_pt->speed = speed;
		    nexthop_pt->x = x;
		    nexthop_pt->y = y;
		}else {
		    new_pt = new nexthop_entry;
		    new_pt->dist = distance;
		    new_pt->next_node = next_hop;
		    new_pt->lastupdate_time = now;
		    new_pt->speed = speed;
		    new_pt->x = x;
		    new_pt->y = y;
		    //if distance is 1, insert in front
		    if (distance == 1) {
			new_pt->next = rt_1[i].next_hop;
			if (rt_1[i].next_hop) {
			    rt_1[i].next_hop->prev = new_pt;
			}
			rt_1[i].next_hop = new_pt;
			new_pt->prev = NULL;
		    }else {
			assert(nexthop_pt->next == NULL);
			//insert at the end of list
			nexthop_pt->next = new_pt;
			new_pt->prev = nexthop_pt;
			new_pt->next = NULL;
		    }
		}
	    }
	    return;
	}else if ((freeslot ==-1) && ((now - rt_1[i].lastupdate_time)>NEIGHBOR_TIMEOUT_INTEVAL)) {
	    //remove all the outdated next hop nodes
	    nexthop_pt = rt_1[i].next_hop;
	    while(nexthop_pt) {
		nexthop_pt = nexthop_pt->next;
		if (nexthop_pt) {
		    delete nexthop_pt->prev;
		}
	    }
	    rt_1[i].next_hop = NULL;
	    freeslot = i;
	}
    }
    assert(freeslot >= 0);

    new_pt = new nexthop_entry;
    new_pt->lastupdate_time = now;
    new_pt->speed = speed;
    new_pt->next_node = next_hop;
    new_pt->x = x;
    new_pt->y = y;
    new_pt->dist = distance;
    new_pt->prev = NULL;
    new_pt->next = NULL;
    rt_1[freeslot].next_hop = new_pt;
    rt_1[freeslot].loc = loc;
    rt_1[freeslot].lastupdate_time = now;
}



nsaddr_t
HGPSRouter::getBestNextHopNode(route_entry &rnt)
{
    nexthop_entry *pt;
    pt = getNextHopNode(rnt.next_hop);
    if (!pt) {
	assert(rnt.next_hop == NULL);
	rnt.dst = 0;
	rnt.lastupdate_time = -3 * NEIGHBOR_TIMEOUT_INTEVAL;
	return 0;
    }
    int dstdst_x,dstdst_y;
    loctodistance(dstdst_x,dstdst_y,rnt.loc);
    dstdst_x = dstdst_x*125;
    dstdst_y = dstdst_y*125;
    double best_dist = 5*RADIO_PROPAGATION*RADIO_PROPAGATION;
    double dist;
    nsaddr_t best_node = pt->next_node;
    while (pt) {
	dist = (pt->x-dstdst_x)*(pt->x-dstdst_x)+(pt->y-dstdst_y)*(pt->y-dstdst_y);
	if (dist<best_dist) {
	    best_node = pt->next_node;
	    best_dist = dist;
	}
	pt = getNextHopNode(pt->next);
    }
    return best_node;
}

bool
HGPSRouter::outdatedNeighbor(nexthop_entry *next_hop, double now)
{
    if ((now - next_hop->lastupdate_time) < NEIGHBOR_TIMEOUT_INTEVAL) {
	return false;
    }else {
	return true;
    }
	
    /*
    double x, y, dist;
    locserver->getPos(x,y);
    dist = (x - next_hop->x)*(x - next_hop->x) + (y-next_hop->y)*(y-next_hop->y);
    dist = sqrt(dist);
    dist += next_hop->speed*(now-next_hop->lastupdate_time);
    if (dist < (0.97*RADIO_PROPAGATION)) {
	return false;
    }else {
	return true;
    }
    */
}

nsaddr_t
HGPSRouter::getNextHopNode(route_entry &rnt)
{
    nexthop_entry *pt;
    pt = getNextHopNode(rnt.next_hop);
    if (!pt) {
	assert(rnt.next_hop == NULL);
	rnt.dst = 0;
	rnt.lastupdate_time = -3*NEIGHBOR_TIMEOUT_INTEVAL;
	return 0;
    }
    return pt->next_node;

}

nexthop_entry *
HGPSRouter::getNextHopNode(nexthop_entry *&next_hop)
{

    nexthop_entry *nexthop_pt;
    nexthop_entry *tmp_pt;

    double now = Scheduler::instance().clock();

    nexthop_pt = next_hop;

    while(nexthop_pt && outdatedNeighbor(nexthop_pt, now)) {
	//delete one entry
	if (nexthop_pt->prev){
	    nexthop_pt->prev->next = nexthop_pt->next;
	    if (nexthop_pt->next) {
	       nexthop_pt->next->prev = nexthop_pt->prev;
	    }
	}else{
	    next_hop = nexthop_pt->next;
	    if (nexthop_pt->next) {
		nexthop_pt->next->prev = NULL;
	    }
	}
	tmp_pt = nexthop_pt;
	nexthop_pt = nexthop_pt->next;
	delete tmp_pt;
    }
    if (nexthop_pt) {
	assert(next_hop !=NULL);
	return nexthop_pt;
    }else{
	return NULL;
    }
}

nsaddr_t 
HGPSRouter::getNextHopFromNT(nsaddr_t dst, location & hisloc)
{
    int i;
    for (int j=0;j<MAX_RT_ENTRIES;j++) {
	i = (j+dst) % MAX_RT_ENTRIES;//some simple hash functions
	if (dst == nt[i].dst) {
	    //if (hisloc.loc != 0) {
		hisloc = nt[i].loc;
	    //}
	    //return getBestNextHopNode(nt[i]);
	    return getNextHopNode(nt[i]);
	}
    }
    return 0;
}

/* this function implements the simplest greedy geographic forwarding
   it sends the packet towards its destination location grid by
   searching in its 1 and 2 hop neighbor table, to prevent from routing
   loop, prev_hop of this route is avoided.  It turns out in the simulation,
   recording only the prev_hop is enough to prevent routing loop for geographic
   forwarding.  More investigation on this issue needs to be done*/
nsaddr_t
HGPSRouter::getNextHop(location hisloc, nsaddr_t prev_hop)
{

    location myloc;
    int dist,min_dist;
    double now = Scheduler::instance().clock();
    nexthop_entry *next_hop_node = NULL;
    int srcdst_x,srcdst_y,dstdst_x,dstdst_y;
    
    myloc = locserver->myloc;
       
    if ((hisloc.mask == BIGGEST_MASK) && (inSameGrid(myloc, hisloc))) {
	//I should be able to reach the destination via my neighbor table
	return parent->myaddr_;
    }

    loctodistance(dstdst_x,dstdst_y,hisloc);
    loctodistance(srcdst_x,srcdst_y,myloc);
    min_dist = (srcdst_x-dstdst_x)*(srcdst_x-dstdst_x)+ (srcdst_y-dstdst_y)*(srcdst_y-dstdst_y);
    
    for (int i=0;i<MAX_ADJ_GRID;i++) {
	if ((now - rt_1[i].lastupdate_time) < NEIGHBOR_TIMEOUT_INTEVAL) {
	    loctodistance(srcdst_x,srcdst_y,rt_1[i].loc);
	    dist = (srcdst_x-dstdst_x)*(srcdst_x-dstdst_x)+ (srcdst_y-dstdst_y)*(srcdst_y-dstdst_y);
	    if (dist < min_dist) {
		if ((next_hop_node = getNextHopNode(rt_1[i].next_hop))) {
		    min_dist = dist;
		}
	    }
	}
    }
    
    if (!next_hop_node) {
	return 0;
    }else {
	//select one that makes the most progress among a set of valid next hops
	nsaddr_t best_next_hop = next_hop_node->next_node;
	double best_dist = (next_hop_node->x-dstdst_x*125)*(next_hop_node->x-dstdst_x*125)+(next_hop_node->y-dstdst_y*125)*(next_hop_node->y-dstdst_y*125);
	double new_dist;
	while (next_hop_node->next) {
	    next_hop_node = getNextHopNode(next_hop_node->next);
	    if (!next_hop_node) return best_next_hop;
	    new_dist = (next_hop_node->x-dstdst_x*125)*(next_hop_node->x-dstdst_x*125)+(next_hop_node->y-dstdst_y*125)*(next_hop_node->y-dstdst_y*125);
	    if ((new_dist < best_dist) && (next_hop_node->next_node!= prev_hop)){
		best_next_hop = next_hop_node->next_node;
		best_dist = new_dist;
	    }
	}
	return best_next_hop;
    }

}

void
HGPSRouter::sendOutPktWithLocation(Packet *&packet, double delay, bool checkNT)
{
    hdr_hgps *hgpsh =  (hdr_hgps*)packet->access(parent->off_hgps_);

    assert(hgpsh->type_ != LOC_QUERY);
    this->sendOutPktWithLocation(packet,delay,checkNT,false);
}

/* invoked by the location server or data packets to send out a packet with its destination grid known.*/
void
HGPSRouter::sendOutPktWithLocation(Packet *&packet,double delay,bool checkNT, bool invoke_loc_server)
{
    hdr_hgps *hgpsh =  (hdr_hgps*)packet->access(parent->off_hgps_);
    hdr_ip *iph =  (hdr_ip*)packet->access(parent->off_ip_);
    hdr_cmn *cmnh =  (hdr_cmn*)packet->access(parent->off_cmn_);
    Scheduler & s = Scheduler::instance();


    //insert failure handler to the packet
    cmnh->xmit_failure_ = HGPS_XmitFailureCallback;
    cmnh->xmit_failure_data_ = (void *) this;

    //check first if there's short-cut
    bool i_have_sent_before = false;

    if (checkNT) {

	for (int i = 0; i<hgpsh->prev_hop_num_; i++) {
	    deleteEntry(hgpsh->prev_hop_[i], iph->dst());
	    if (hgpsh->prev_hop_[i] == parent->myaddr_) {
		i_have_sent_before = true;
	    }
	}
	
	if ((iph->dst_!=0) && (cmnh->next_hop_ = getNextHopFromNT(iph->dst_,hgpsh->dst_loc_))) {
	    if (cmnh->next_hop_!=iph->dst()) {
		if (hgpsh->prev_hop_num_ >= 4) {
		    parent->drop(packet, DROP_RTR_ROUTE_LOOP);
		    return;
		}else{
		    if (hgpsh->prev_hop_[hgpsh->prev_hop_num_-1]!=parent->myaddr_) {
		    	hgpsh->prev_hop_[hgpsh->prev_hop_num_] = parent->myaddr_;
		    	hgpsh->prev_hop_num_++;
		    }
		}
	    }
    	    s.schedule(parent->ll,packet,0);
	    packet = NULL;
    	    return;
	}
    }
    assert((hgpsh->dst_loc_).mask != 0);

#ifdef IDEAL_ROUTE 
    //this part has not been tested!
    if (hgpsh->curr_addr_ == 0) {
	//i'll ask god to give me a perfect route!
	if (God::instance()->getRouteToLoc(parent->myaddr_,hgpsh->dst_loc_,hgpsh->addrs_,MAX_ADDR_LEN)) {
	    parent->drop(packet,DROP_RTR_NO_ROUTE);
	}else {
	    hgpsh->curr_addr_++;
	    cmnh->next_hop() = hgpsh->addrs_[hgpsh->curr_addr_];
	    s.schedule(parent->ll,packet,delay);
	}
    }else {
	//just forward the pkt along the source route
       hgpsh->curr_addr_++;
       cmnh->next_hop() = hgpsh->addrs_[hgpsh->curr_addr_];
       s.schedule(parent->ll,packet,delay);
    }
    packet = NULL;
    return;
#endif
    
    cmnh->next_hop_ = getNextHop(hgpsh->dst_loc_, (hgpsh->prev_hop_num_)? hgpsh->prev_hop_[0]:0);
    /* an location update might not have the idea of who to forward the pkt to*/
    /* i have to set iph->dst(), otherwise the link layer will complain*/
    if (!iph->dst()) {
	iph->dst() = cmnh->next_hop(); 
    }
    
    if (cmnh->next_hop() && (cmnh->next_hop() != parent->myaddr_)) {
	hgpsh->prev_hop_[0] = parent->myaddr_;
	hgpsh->prev_hop_num_ = 1;
	s.schedule(parent->ll,packet,delay);
    }else if (cmnh->next_hop() == parent->myaddr_) {
	//I should be the next hop to destination. i.e. destination has moved away
	//send out an ERR packet to information the source about this invalid dst_loc_
#ifdef HGPS_DEBUG
	printf("_%d_ %.2f location (%d) error dst %d\n",parent->myaddr_,s.clock(),hgpsh->dst_loc_.loc,iph->dst());
#endif
	if (hgpsh->type_ == DATA) {
    	    Packet *p = parent->allocpkt();
	    hdr_hgps *new_hgpsh =  (hdr_hgps*)p->access(parent->off_hgps_);
	    hdr_ip *new_iph =  (hdr_ip*)p->access(parent->off_ip_);
	    hdr_cmn *new_cmnh =  (hdr_cmn*)p->access(parent->off_cmn_);
       
	    new_iph->src() = parent->myaddr_;
	    new_iph->dst() = iph->src();
	    new_hgpsh->init();
	    new_hgpsh->dst_loc_ = hgpsh->src_loc_;
	    new_hgpsh->src_loc_ = locserver->getGrid();
	    new_hgpsh->type_ = ERR;
	    new_hgpsh->id_ = iph->dst();
	    new_hgpsh->loc_ = hgpsh->dst_loc_;

	    new_cmnh->num_forwards() = 0;
	    new_cmnh->ptype() = PT_HGPS;
	    new_cmnh->size() = hgpsh->size();

	    sendOutPktWithLocation(p,0.0,true);
    	    parent->trace("HGPS DROP: _%d_ %.2f outdated location %d->%d(%d)",parent->myaddr_,s.clock(),iph->src(),iph->dst(),(hgpsh->dst_loc_).loc);

	}else if ((hgpsh->type_ == LOC_QUERY)) {
	    if (parent->verbose_) {
		parent->trace("HGPS dst %d loc %d error",iph->dst(),hgpsh->dst_loc_.loc);
	    }

	    if (invoke_loc_server) {
		locserver->recvQuery(packet,true);
		return;
	    }else{
		parent->trace("HGPS %.5f _%d_ loc-query-failure-1(%d %d)%d->%d",s.clock(),parent->myaddr_,parent->myaddr_,hgpsh->id_,iph->src(),iph->dst());
	    }
	}else if (hgpsh->type_ == LOC_UPDATE) {
	    printf("_%d_ update dst %d location(%d) error\n", parent->myaddr_,iph->dst(),hgpsh->dst_loc_.loc);
	}
	//i classify location error as no route temporarily
	parent->drop(packet,DROP_RTR_NO_ROUTE);

    }else{
	double x, y, z;
	locserver->mynode->getLoc(&x,&y,&z);
#ifdef HGPS_DEBUG
	printf("_%d_ (%d, %.2f, %.2f) %.5f no route to dst %d (%d,%d)\n",parent->myaddr_,
		(locserver->getGrid()).loc,x,y,s.clock(),iph->dst(),(hgpsh->dst_loc_).loc,(hgpsh->dst_loc_).mask);
#endif
	parent->drop(packet,DROP_RTR_NO_ROUTE);
	packet = NULL;
	return;
    }
    packet = NULL;
}

void
HGPS_XmitFailureCallback(Packet *pkt, void *data)
{
    HGPSRouter *router = (HGPSRouter *)data; // cast of trust
    router->xmitFailed(pkt);
}

void
HGPSRouter::xmitFailed(Packet *pkt)
{
    hdr_hgps *hgpsh =  (hdr_hgps*)pkt->access(parent->off_hgps_);
    hdr_ip *iph =  (hdr_ip*)pkt->access(parent->off_ip_);
    hdr_cmn *cmnh =  (hdr_cmn*)pkt->access(parent->off_cmn_);
    
    double now = Scheduler::instance().clock();


    //delete incorrect neighbor entry whatever
    deleteEntry(cmnh->next_hop());
    
    if (parent->verbose_) {
	parent->trace("HGPS: error from %d->%d %d",parent->myaddr_,cmnh->next_hop(),cmnh->uid());
    }

    nsaddr_t prev_cmnh_next_hop = cmnh->next_hop();
    cmnh->next_hop() = 0; //i have to reset next hop
/*
    if (hgpsh->prev_hop_num_ && (hgpsh->prev_hop_[hgpsh->prev_hop_num_-1] == parent->myaddr_)) {
	    hgpsh->prev_hop_num_ --;
	}
*/

    if (hgpsh->type_ == LOC_UPDATE) {
	//because my neighbor table has been changed, therefore iph->dst() might be different from recvUpdate()
        locserver->recvUpdate(pkt);     
    }else if (hgpsh->type_ == LOC_QUERY) {
        locserver->recvQuery(pkt,false);
    }else {
        sendOutPktWithLocation(pkt,0.0,true);
    }

    /* now kill all the other packets in the output queue that would
    use the same next hop.  This is reasonable, since 802.11 has
    already retried the xmission multiple times => a persistent failure. 
    from dmaltz*/

    Packet *r, *nr, *head = NULL;		// pkts to be recycled
    while ((r = (parent->ifq)->prq_get_nexthop(prev_cmnh_next_hop))) {
	r->next_ = head;
	head = r; 
    }

    // the packets are now in the reverse order of how they appeared in the IFQ
    // so reverse them again
    for (r = head, head = NULL; r; r = r->next_) {  
	r->next_ = head;
	head = r; 
    }

    // now process them in order
    for (r = head; r; r = nr) {  
	nr = r->next_;

	//depending on packet type, rescue differently
	hgpsh =  (hdr_hgps*)r->access(parent->off_hgps_);
	cmnh = (hdr_cmn*)r->access(parent->off_cmn_);

	cmnh->next_hop() = 0;
	/*
	if (hgpsh->prev_hop_num_ && (hgpsh->prev_hop_[hgpsh->prev_hop_num_-1] == parent->myaddr_)) {
	    hgpsh->prev_hop_num_ --;
	}
	*/

	if (hgpsh->type_ == LOC_UPDATE) {
	    locserver->recvUpdate(r);
	}else if (hgpsh->type_ == LOC_QUERY) {
	    locserver->recvQuery(r,false);
	}else {
	    sendOutPktWithLocation(r,jitter(0.2,1),true);
	}
    } 
}
