/* mac-802_11.cc
   $Id: mac-802_11.cc,v 1.1.1.1 2000/08/28 18:40:06 jinyang Exp $
   */

#include <delay.h>
#include <connector.h>
#include <packet.h>
#include <random.h>

// #define DEBUG

#include <cmu/debug.h>
#include <cmu/arp.h>
#include <cmu/ll.h> 
#include <cmu/mac.h>
#include <cmu/mac-timers.h>
#include <cmu/mac-802_11.h>
#include <cmu/cmu-trace.h>


/* ======================================================================
   Macros
   ====================================================================== */
#define CHECK_BACKOFF_TIMER()						\
{									\
	if(is_idle() && mhBackoff.paused())				\
		mhBackoff.resume(difs);					\
	if(! is_idle() && mhBackoff.busy() && ! mhBackoff.paused())	\
		mhBackoff.pause();					\
}

#define TRANSMIT(p, t)                                                  \
{                                                                       \
	tx_active = 1;                                                  \
                                                                        \
        /*                                                              \
         * If I'm transmitting without doing CS, such as when           \
         * sending an ACK, any incoming packet will be "missed"         \
         * and hence, must be discarded.                                \
         */                                                             \
        if(rx_state != MAC_IDLE) {                                      \
                struct hdr_mac802_11 *dh = HDR_MAC11(p);                \
                                                                        \
                assert(dh->dh_fc.fc_type == MAC_Type_Control);          \
                assert(dh->dh_fc.fc_subtype == MAC_Subtype_ACK);        \
                                                                        \
                assert(pktRx);                                          \
                struct hdr_cmn *ch = HDR_CMN(pktRx);                    \
                                                                        \
                ch->error() = 1;        /* force packet discard */      \
        }                                                               \
                                                                        \
        /*                                                              \
         * pass the packet on the "interface" which will in turn        \
         * place the packet on the channel.                             \
         *                                                              \
         * NOTE: a handler is passed along so that the Network          \
         *       Interface can distinguish between incoming and         \
         *       outgoing packets.                                      \
         */                                                             \
        sendtarget_->recv(p->copy(), this);                             \
                                                                        \
        mhSend.start(t);                                                \
                                                                        \
	mhIF.start(TX_Time(p));                                         \
}

#define SET_RX_STATE(x)			\
{					\
	rx_state = (x);			\
					\
	CHECK_BACKOFF_TIMER();		\
}

#define SET_TX_STATE(x)				\
{						\
	tx_state = (x);				\
						\
	CHECK_BACKOFF_TIMER();			\
}


/* ======================================================================
   Global Variables
   ====================================================================== */
extern char* pt_names[];

static PHY_MIB PMIB =
{
	DSSS_CWMin, DSSS_CWMax, DSSS_SlotTime, DSSS_CCATime,
	DSSS_RxTxTurnaroundTime, DSSS_SIFSTime, DSSS_PreambleLength,
	DSSS_PLCPHeaderLength
};	

static MAC_MIB MMIB =
{
	0 /* MAC_RTSThreshold */, MAC_ShortRetryLimit,
	MAC_LongRetryLimit, MAC_FragmentationThreshold,
	MAC_MaxTransmitMSDULifetime, MAC_MaxReceiveLifetime,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};



/* ======================================================================
   TCL Hooks for the simulator
   ====================================================================== */
static class Mac802_11Class : public TclClass {
public:
	Mac802_11Class() : TclClass("Mac/802_11") {}
	TclObject* create(int, const char*const*) {
		return (new Mac802_11(&PMIB, &MMIB));
	}
} class_mac802_11;


/* ======================================================================
   Mac Class Functions
   ====================================================================== */
Mac802_11::Mac802_11(PHY_MIB *p, MAC_MIB *m) : Mac(), initialized_(0),
	mhIF(this), mhNav(this), mhRecv(this), mhSend(this), 
	mhDefer(this, p->SlotTime), mhBackoff(this, p->SlotTime)
{
	macmib = m;
	phymib = p;

	nav = 0.0;

	tx_state = rx_state = MAC_IDLE;
	tx_active = 0;

	pktRTS = 0;
	pktCTRL = 0;

	cw = phymib->CWMin;
	ssrc = slrc = 0;

	sta_seqno = 1;
	cache = 0;
	cache_node_count = 0;
}
void
Mac802_11::init() 
{
	Mac::init();

	sifs = phymib->SIFSTime;
	pifs = sifs + phymib->SlotTime;
	difs = sifs + 2*phymib->SlotTime;
	eifs = sifs + difs + DATA_Time(ETHER_ACK_LEN +
		phymib->PreambleLength/8 + phymib->PLCPHeaderLength/8);
	tx_sifs = sifs - phymib->RxTxTurnaroundTime;
	tx_pifs = tx_sifs + phymib->SlotTime;
	tx_difs = tx_sifs + 2 * phymib->SlotTime;
	initialized_ = 1;
}

int
Mac802_11::command(int argc, const char*const* argv)
{
	if (argc == 2) {
		if(strcasecmp(argv[1], "initialize") == 0) {
			init();
			return TCL_OK;
		}
	} else if (argc == 3) {
		
		if (strcmp(argv[1], "log-target") == 0) {
			logtarget_ = (NsObject*) TclObject::lookup(argv[2]);
			if(logtarget_ == 0)
				return TCL_ERROR;
			return TCL_OK;
		}

		if(strcmp(argv[1], "nodes") == 0) {
			if(cache) return TCL_ERROR;
			cache_node_count = atoi(argv[2]);
			cache = new Host[cache_node_count + 1];
			assert(cache);
			bzero(cache, sizeof(Host) * (cache_node_count+1 ));
			return TCL_OK;
		}
	}
	return Mac::command(argc, argv);
}


/* ======================================================================
   Debugging Routines
   ====================================================================== */
void
Mac802_11::trace_pkt(Packet *p) {
	struct hdr_cmn *ch = HDR_CMN(p);
	struct hdr_mac802_11* dh = HDR_MAC11(p);
	u_int16_t *t = (u_int16_t*) &dh->dh_fc;

	fprintf(stderr, "\t[ %2x %2x %2x %2x ] %x %s %d\n",
		*t, dh->dh_duration,
		ETHER_ADDR(dh->dh_da), ETHER_ADDR(dh->dh_sa),
		index, pt_names[ch->ptype()], ch->size());
}

void
Mac802_11::dump(char *fname)
{
	fprintf(stderr,
		"\n%s --- (INDEX: %d, time: %2.9f)\n",
		fname, index, Scheduler::instance().clock());

	fprintf(stderr,
		"\ttx_state: %x, rx_state: %x, nav: %2.9f, idle: %d\n",
		tx_state, rx_state, nav, is_idle());

	fprintf(stderr,
		"\tpktTx: %x, pktRx: %x, pktRTS: %x, pktCTRL: %x, upcall: %x\n",
		(int) pktTx, (int) pktRx, (int) pktRTS,
		(int) pktCTRL, (int) upcall_);

	fprintf(stderr,
		"\tDefer: %d, Backoff: %d (%d), Recv: %d, Timer: %d Nav: %d\n",
		mhDefer.busy(), mhBackoff.busy(), mhBackoff.paused(),
		mhRecv.busy(), mhSend.busy(), mhNav.busy());
	fprintf(stderr,
		"\tBackoff Expire: %f\n",
		mhBackoff.expire());
}


/* ======================================================================
   Packet Headers Routines
   ====================================================================== */
inline MacAddr
Mac802_11::hdr_dst(void* hdr, MacAddr dst)
{
	struct hdr_mac802_11 *dh = (struct hdr_mac802_11*) hdr;

	if(dst)
		STORE4BYTE(&dst,(dh->dh_da));

	return ETHER_ADDR(dh->dh_da);
}

inline MacAddr 
Mac802_11::hdr_src(void* hdr, MacAddr src)
{
	struct hdr_mac802_11 *dh = (struct hdr_mac802_11*) hdr;

	if(src)
		STORE4BYTE(&src,dh->dh_sa);

	return ETHER_ADDR(dh->dh_sa);
}

inline int 
Mac802_11::hdr_type(void* hdr, u_int16_t type)
{
	struct hdr_mac802_11 *dh = (struct hdr_mac802_11*) hdr;

	if(type)
	        SET_ETHER_TYPE(dh->dh_body, type);

	return GET_ETHER_TYPE(dh->dh_body);
}


/* ======================================================================
   Misc Routines
   ====================================================================== */
inline int
Mac802_11::is_idle()
{
	if(rx_state != MAC_IDLE)
		return 0;

	if(tx_state != MAC_IDLE)
		return 0;

	if(nav > Scheduler::instance().clock())
		return 0;

	return 1;
}


void
Mac802_11::discard(Packet *p, const char* why)
{
	hdr_mac802_11* mh = HDR_MAC11(p);
	hdr_cmn *ch = HDR_CMN(p);

#if 0
	/* old logic 8/8/98 -dam */
	/*
	 * If received below the RXThreshold, then just free.
	 */
	if(p->txinfo.Pr < p->txinfo.ant.RXThresh) {
		Packet::free(p);
		return;
	}
#endif 0

	/* if the rcvd pkt contains errors, a real MAC layer couldn't
	   necessarily read any data from it, so we just toss it now */
	if(ch->error() != 0) {
		Packet::free(p);
		return;
	}

	switch(mh->dh_fc.fc_type) {

	case MAC_Type_Management:
		drop(p, why);
		return;

	case MAC_Type_Control:
		switch(mh->dh_fc.fc_subtype) {

		case MAC_Subtype_RTS:
			if(ETHER_ADDR(mh->dh_sa) == index) {
				drop(p, why);
				return;
			}

		/* fall through - if necessary */

		case MAC_Subtype_CTS:
		case MAC_Subtype_ACK:
			if(ETHER_ADDR(mh->dh_da) == index) {
				drop(p, why);
				return;
			}
			break;

		default:
			fprintf(stderr, "invalid MAC Control subtype\n");
			exit(1);
		}
		break;

	case MAC_Type_Data:
		switch(mh->dh_fc.fc_subtype) {

		case MAC_Subtype_Data:
			if(ETHER_ADDR(mh->dh_da) == index ||
			   ETHER_ADDR(mh->dh_sa) == index ||
			   ETHER_ADDR(mh->dh_da) == MAC_BROADCAST) {
				drop(p, why);
				return;
			}
			break;

		default:
			fprintf(stderr, "invalid MAC Data subtype\n");
			exit(1);
		}
		break;

	default:
		fprintf(stderr, "invalid MAC type (%x)\n", mh->dh_fc.fc_type);
		trace_pkt(p);
		exit(1);
	}
	Packet::free(p);
}


void
Mac802_11::capture(Packet *p)
{
	/*
	 * Update the NAV so that this does not screw
	 * up carrier sense.
	 */
	set_nav(usec(eifs + TX_Time(p)));

	Packet::free(p);
}


void
Mac802_11::collision(Packet *p)
{
	switch(rx_state) {

	case MAC_RECV:
		SET_RX_STATE(MAC_COLL);

		/* fall through */

	case MAC_COLL:
		assert(pktRx);
		assert(mhRecv.busy());

		/*
		 *  Since a collision has occurred, figure out
		 *  which packet that caused the collision will
		 *  "last" the longest.  Make this packet,
		 *  pktRx and reset the Recv Timer if necessary.
		 */
		if(TX_Time(p) > mhRecv.expire()) {
			mhRecv.stop();
			discard(pktRx, DROP_MAC_COLLISION);
			pktRx = p;
			mhRecv.start(TX_Time(pktRx));
		}
		else {
			discard(p, DROP_MAC_COLLISION);
		}
		break;

	default:
		assert(0);
	}
}


void
Mac802_11::tx_resume()
{
	assert(mhSend.busy() == 0);
	assert(mhDefer.busy() == 0);

	if(pktCTRL) {
		/*
		 *  Need to send a CTS or ACK.
		 */
		mhDefer.start(sifs);
	}
	else if(pktRTS) {

		if(mhBackoff.busy() == 0)
			mhDefer.start(difs);

	}
	else if(pktTx) {

		if(mhBackoff.busy() == 0)
			mhDefer.start(difs);

	}
	else if(upcall_) {
		Handler *h = upcall_;
		upcall_ = 0;
		h->handle((Event*) 0);
	}

	SET_TX_STATE(MAC_IDLE);
}


void
Mac802_11::rx_resume()
{
	assert(pktRx == 0);
	assert(mhRecv.busy() == 0);

	SET_RX_STATE(MAC_IDLE);
}


/* ======================================================================
   Timer Handler Routines
   ====================================================================== */
void
Mac802_11::backoffHandler()
{
	if(pktCTRL) {
		assert(mhSend.busy() || mhDefer.busy());
		return;
	}

	if(check_pktRTS() == 0)
		return;

	if(check_pktTx() == 0)
		return;
}

void
Mac802_11::deferHandler()
{
	assert(pktCTRL || pktRTS || pktTx);

	if(check_pktCTRL() == 0)
		return;

	assert(mhBackoff.busy() == 0);

	if(check_pktRTS() == 0)
		return;

	if(check_pktTx() == 0)
		return;
}

void
Mac802_11::navHandler()
{
	if(is_idle() && mhBackoff.paused())
		mhBackoff.resume(difs);
}

void
Mac802_11::recvHandler()
{
	recv_timer();
}

void
Mac802_11::sendHandler()
{
	send_timer();
}


void
Mac802_11::txHandler()
{
	tx_active = 0;
}


/* ======================================================================
   The "real" Timer Handler Routines
   ====================================================================== */
void
Mac802_11::send_timer()
{

	switch(tx_state) {

	/*
	 * Sent a RTS, but did not receive a CTS.
	 */
	case MAC_RTS:
		RetransmitRTS();

		break;

	/*
	 * Sent a CTS, but did not receive a DATA packet.
	 */
	case MAC_CTS:
		assert(pktCTRL);
		Packet::free(pktCTRL); pktCTRL = 0;

		break;

	/*
	 * Sent DATA, but did not receive an ACK packet.
	 */
	case MAC_SEND:
		RetransmitDATA();

		break;
		
	/*
	 * Sent an ACK, and now ready to resume transmission.
	 */
	case MAC_ACK:
		assert(pktCTRL);
		Packet::free(pktCTRL); pktCTRL = 0;
		break;


	case MAC_IDLE:
		break;

	default:
		assert(0);

	}

	tx_resume();
}


/* ======================================================================
   Outgoing Packet Routines
   ====================================================================== */
int
Mac802_11::check_pktCTRL()
{
	struct hdr_mac802_11 *mh;
	double timeout;

	if(pktCTRL == 0)
		return -1;
	if(tx_state == MAC_CTS || tx_state == MAC_ACK)
		return -1;

	mh = HDR_MAC11(pktCTRL);

	switch(mh->dh_fc.fc_subtype) {

	/*
	 *  If the medium is not IDLE, don't send the CTS.
	 */
	case MAC_Subtype_CTS:
		if(! is_idle()) {
			discard(pktCTRL, DROP_MAC_BUSY); pktCTRL = 0;
			return 0;
		}

		SET_TX_STATE(MAC_CTS);

		timeout = (mh->dh_duration * 1e-6) + CTS_Time; // XXX

		break;

	/*
	 * IEEE 802.11 specs, section 9.2.8
	 * Acknowledments are sent after an SIFS, without regard to
	 * the busy/idle state of the medium.
	 */
	case MAC_Subtype_ACK:
		SET_TX_STATE(MAC_ACK);
		timeout = ACK_Time;

		break;

	default:
		fprintf(stderr, "Invalid MAC Control subtype\n");
		exit(1);
	}

        TRANSMIT(pktCTRL, timeout);

	return 0;
}


int
Mac802_11::check_pktRTS()
{
	struct hdr_mac802_11 *mh;
	double timeout;

	assert(mhBackoff.busy() == 0);

	if(pktRTS == 0)
		return -1;

	mh = HDR_MAC11(pktRTS);

	switch(mh->dh_fc.fc_subtype) {

	case MAC_Subtype_RTS:
		if(! is_idle()) {
			inc_cw();
			mhBackoff.start(cw, is_idle());
			return 0;
		}

		SET_TX_STATE(MAC_RTS);
		timeout = CTSTimeout;

		break;

	default:
		fprintf(stderr, "Invalid MAC Control subtype\n");
		exit(1);
	}

        TRANSMIT(pktRTS, timeout);

	return 0;
}


int
Mac802_11::check_pktTx()
{
	struct hdr_mac802_11 *mh;
	double timeout;

	assert(mhBackoff.busy() == 0);

	if(pktTx == 0)
		return -1;

	mh = HDR_MAC11(pktTx);
       	int len = HDR_CMN(pktTx)->size();

	switch(mh->dh_fc.fc_subtype) {

	case MAC_Subtype_Data:

		if(! is_idle()) {
			sendRTS(ETHER_ADDR(mh->dh_da));

			inc_cw();
			mhBackoff.start(cw, is_idle());
			return 0;
		}

		SET_TX_STATE(MAC_SEND);

		if(ETHER_ADDR(mh->dh_da) != MAC_BROADCAST)
			timeout = ACKTimeout(len);
		else
			timeout = TX_Time(pktTx);

		break;

	default:
		fprintf(stderr, "Invalid MAC Control subtype\n");
		exit(1);
	}

        TRANSMIT(pktTx, timeout);

	return 0;
}


/*
 * Low-level transmit functions that actually place the packet onto
 * the channel.
 */
void
Mac802_11::sendRTS(int dst)
{
	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct rts_frame *rf = (struct rts_frame*)p->access(off_mac_);

	assert(pktTx);
	assert(pktRTS == 0);

	/*
	 *  If the size of the packet is larger than the
	 *  RTSThreshold, then perform the RTS/CTS exchange.
	 *
	 *  XXX: also skip if destination is a broadcast
	 */
	if( (u_int32_t) HDR_CMN(pktTx)->size() < macmib->RTSThreshold ||
	    (u_int32_t) dst == MAC_BROADCAST) {
		Packet::free(p);
		return;
	}

	ch->uid() = 0;
	ch->ptype() = PT_MAC;
	ch->size() = ETHER_RTS_LEN;
	ch->iface() = -2;
	ch->error() = 0;

	bzero(rf, MAC_HDR_LEN);

	rf->rf_fc.fc_protocol_version = MAC_ProtocolVersion;
	rf->rf_fc.fc_type	= MAC_Type_Control;
	rf->rf_fc.fc_subtype	= MAC_Subtype_RTS;
	rf->rf_fc.fc_to_ds	= 0;
	rf->rf_fc.fc_from_ds	= 0;
	rf->rf_fc.fc_more_frag	= 0;
	rf->rf_fc.fc_retry	= 0;
	rf->rf_fc.fc_pwr_mgt	= 0;
	rf->rf_fc.fc_more_data	= 0;
	rf->rf_fc.fc_wep	= 0;
	rf->rf_fc.fc_order	= 0;

	rf->rf_duration = RTS_DURATION(pktTx);
	STORE4BYTE(&dst,rf->rf_ra);
	STORE4BYTE(&index,rf->rf_ta);
	// rf->rf_fcs;

	pktRTS = p;
}

void
Mac802_11::sendCTS(int dst, double rts_duration)
{
	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct cts_frame *cf = (struct cts_frame*)p->access(off_mac_);

	assert(pktCTRL == 0);

	ch->uid() = 0;
	ch->ptype() = PT_MAC;
	ch->size() = ETHER_CTS_LEN;
	ch->iface() = -2;
	ch->error() = 0;

	bzero(cf, MAC_HDR_LEN);

	cf->cf_fc.fc_protocol_version = MAC_ProtocolVersion;
	cf->cf_fc.fc_type	= MAC_Type_Control;
	cf->cf_fc.fc_subtype	= MAC_Subtype_CTS;
	cf->cf_fc.fc_to_ds	= 0;
	cf->cf_fc.fc_from_ds	= 0;
	cf->cf_fc.fc_more_frag	= 0;
	cf->cf_fc.fc_retry	= 0;
	cf->cf_fc.fc_pwr_mgt	= 0;
	cf->cf_fc.fc_more_data	= 0;
	cf->cf_fc.fc_wep	= 0;
	cf->cf_fc.fc_order	= 0;

	cf->cf_duration = CTS_DURATION(rts_duration);
	STORE4BYTE(&dst,cf->cf_ra);
	// cf->cf_fcs;

	pktCTRL = p;
}

void
Mac802_11::sendACK(int dst)
{
	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct ack_frame *af = (struct ack_frame*)p->access(off_mac_);

	assert(pktCTRL == 0);

	ch->uid() = 0;
	ch->ptype() = PT_MAC;
	ch->size() = ETHER_ACK_LEN;
	ch->iface() = -2;
	ch->error() = 0;

	bzero(af, MAC_HDR_LEN);

	af->af_fc.fc_protocol_version = MAC_ProtocolVersion;
	af->af_fc.fc_type	= MAC_Type_Control;
	af->af_fc.fc_subtype	= MAC_Subtype_ACK;
	af->af_fc.fc_to_ds	= 0;
	af->af_fc.fc_from_ds	= 0;
	af->af_fc.fc_more_frag	= 0;
	af->af_fc.fc_retry	= 0;
	af->af_fc.fc_pwr_mgt	= 0;
	af->af_fc.fc_more_data	= 0;
	af->af_fc.fc_wep	= 0;
	af->af_fc.fc_order	= 0;

	af->af_duration = ACK_DURATION();
	STORE4BYTE(&dst,af->af_ra);
	// af->af_fcs;

	pktCTRL = p;
}

void
Mac802_11::sendDATA(Packet *p)
{
	hdr_cmn* ch = HDR_CMN(p);
	struct hdr_mac802_11* dh = HDR_MAC11(p);

	assert(pktTx == 0);

	/*
	 * Update the MAC header
	 */
	ch->size() += ETHER_HDR_LEN11;

	dh->dh_fc.fc_protocol_version = MAC_ProtocolVersion;
	dh->dh_fc.fc_type       = MAC_Type_Data;
	dh->dh_fc.fc_subtype    = MAC_Subtype_Data;
	dh->dh_fc.fc_to_ds      = 0;
	dh->dh_fc.fc_from_ds    = 0;
	dh->dh_fc.fc_more_frag  = 0;
	dh->dh_fc.fc_retry      = 0;
	dh->dh_fc.fc_pwr_mgt    = 0;
	dh->dh_fc.fc_more_data  = 0;
	dh->dh_fc.fc_wep        = 0;
	dh->dh_fc.fc_order      = 0;

	if(ETHER_ADDR(dh->dh_da) != MAC_BROADCAST)
		dh->dh_duration = DATA_DURATION();
	else
		dh->dh_duration = 0;

	pktTx = p;
}


/* ======================================================================
   Retransmission Routines
   ====================================================================== */
void
Mac802_11::RetransmitRTS()
{
	assert(pktTx);
	assert(pktRTS);

	assert(mhBackoff.busy() == 0);

	macmib->RTSFailureCount++;

	ssrc += 1;			// STA Short Retry Count

	if(ssrc >= macmib->ShortRetryLimit) {

		discard(pktRTS, DROP_MAC_RETRY_COUNT_EXCEEDED); pktRTS = 0;

		/* tell the callback the send operation failed 
		   before discarding the packet */
		hdr_cmn *ch = HDR_CMN(pktTx);
		if (ch->xmit_failure_) {
                        /*
                         *  Need to remove the MAC header so that 
                         *  re-cycled packets don't keep getting
                         *  bigger.
                         */
                        ch->size() -= ETHER_HDR_LEN11;
                        ch->xmit_reason_ = XMIT_REASON_RTS;
                        ch->xmit_failure_(pktTx->copy(),
                                          ch->xmit_failure_data_);
                }

		discard(pktTx, DROP_MAC_RETRY_COUNT_EXCEEDED); pktTx = 0;

		ssrc = 0;
		rst_cw();
	}
	else {
		struct rts_frame *rf;

		rf = (struct rts_frame*)pktRTS->access(off_mac_);
		rf->rf_fc.fc_retry = 1;

		inc_cw();
		mhBackoff.start(cw, is_idle());
	}
}


void
Mac802_11::RetransmitDATA()
{
	struct hdr_cmn *ch;
	struct hdr_mac802_11 *mh;
	u_int32_t *rcount, *thresh;

	assert(mhBackoff.busy() == 0);

	assert(pktTx);
	assert(pktRTS == 0);

	ch = HDR_CMN(pktTx);
	mh = HDR_MAC11(pktTx);

	/*
	 *  Broadcast packets don't get ACKed and therefore
	 *  are never retransmitted.
	 */
	if(ETHER_ADDR(mh->dh_da) == MAC_BROADCAST) {
		Packet::free(pktTx); pktTx = 0;

		/*
		 * Backoff at end of TX.
		 */
		rst_cw();
		mhBackoff.start(cw, is_idle());

		return;
	}

	macmib->ACKFailureCount++;

	if((u_int32_t) ch->size() <= macmib->RTSThreshold) {
		rcount = &ssrc;
		thresh = &macmib->ShortRetryLimit;
	}
	else {
		rcount = &slrc;
		thresh = &macmib->LongRetryLimit;
	}

	(*rcount)++;

	if(*rcount > *thresh) {

		macmib->FailedCount++;

		/* tell the callback the send operation failed 
		   before discarding the packet */
		hdr_cmn *ch = HDR_CMN(pktTx);
		if (ch->xmit_failure_) {
                        ch->size() -= ETHER_HDR_LEN11;
                        ch->xmit_reason_ = XMIT_REASON_ACK;
                        ch->xmit_failure_(pktTx->copy(),
                                          ch->xmit_failure_data_);
                }

		discard(pktTx, DROP_MAC_RETRY_COUNT_EXCEEDED); pktTx = 0;

		*rcount = 0;
		rst_cw();
	}
	else {
		struct hdr_mac802_11 *dh;

		dh = HDR_MAC11(pktTx);
		dh->dh_fc.fc_retry = 1;

		sendRTS(ETHER_ADDR(mh->dh_da));
		inc_cw();
		mhBackoff.start(cw, is_idle());
	}
}


/* ======================================================================
   Incoming Packet Routines
   ====================================================================== */
void
Mac802_11::send(Packet *p, Handler *h)
{
	struct hdr_mac802_11* dh = HDR_MAC11(p);
	struct hdr_cmn *hdr = HDR_CMN(p);

	if ((hdr->uid() == 1016) && (index == 82)) {
	    printf("haha\n");
	}

	upcall_ = h;

	sendDATA(p);
	sendRTS(ETHER_ADDR(dh->dh_da));

	/*
	 * Assign the data packet a sequence number.
	 */
	dh->dh_scontrol = sta_seqno++;

	/*
	 *  If the medium is IDLE, we must wait for a DIFS
	 *  Space before transmitting.
	 */
	if(mhBackoff.busy() == 0) {
		if(is_idle()) {
			/*
			 * If we are already deferring, there is no
			 * need to reset the Defer timer.
			 */
			if(mhDefer.busy() == 0)
				mhDefer.start(difs);
		}

		/*
		 * If the medium is NOT IDLE, then we start
		 * the backoff timer.
		 */
		else {
		mhBackoff.start(cw, is_idle());
		}
	}
}


void
Mac802_11::recv(Packet *p, Handler *h)
{
	struct hdr_cmn *hdr = HDR_CMN(p);

	/*
	 * Sanity Check
	 */
	assert(initialized());

	if (asleep) {
	    /* The interface is sleeping, discard the packet */
	    discard(p, "---");
	    return;
	}

	/*
	 *  Handle outgoing packets.
	 */
	if(h) {
                send(p, h);
                return;
        }

	/*
	 *  Handle incoming packets.
	 *
	 *  We just received the 1st bit of a packet on the network
	 *  interface.
	 *
	 */

	/*
	 *  If the interface is currently in transmit mode, then
	 *  it probably won't even see this packet.  However, the
	 *  "air" around me is BUSY so I need to let the packet
	 *  proceed.  Just set the error flag in the common header
	 *  to that the packet gets thrown away.
	 */
	if(tx_active && hdr->error() == 0) {
		hdr->error() = 1;
	}

	if(rx_state == MAC_IDLE) {
		SET_RX_STATE(MAC_RECV);
		pktRx = p;

		/*
		 * Schedule the reception of this packet, in
		 * txtime seconds.
		 */
		mhRecv.start(TX_Time(p));
	}
	else {
		/*
		 *  If the power of the incoming packet is smaller than the
		 *  power of the packet currently being received by at least
                 *  the capture threshold, then we ignore the new packet.
		 */
		if(pktRx->txinfo.RxPr / p->txinfo.RxPr >= p->txinfo.CPThresh) {
			capture(p);
		}
		else {
			collision(p);
		}
	}
}


void
Mac802_11::recv_timer()
{
	hdr_cmn *ch = HDR_CMN(pktRx);
	hdr_mac802_11 *mh = HDR_MAC11(pktRx);
	u_int32_t dst = ETHER_ADDR(mh->dh_da);
	u_int8_t  type = mh->dh_fc.fc_type;
	u_int8_t  subtype = mh->dh_fc.fc_subtype;

	assert(pktRx);
	assert(rx_state == MAC_RECV || rx_state == MAC_COLL);

        /*
         *  If the interface is in TRANSMIT mode when this packet
         *  "arrives", then I would never have seen it and should
         *  do a silent discard without adjusting the NAV.
         */
        if(tx_active) {
                Packet::free(pktRx);
                goto done;
        }

	/*
	 * Handle collisions.
	 */
	if(rx_state == MAC_COLL) {
		discard(pktRx, DROP_MAC_COLLISION);
		set_nav(usec(eifs));
		goto done;
	}

	/*
	 * Check to see if this packet was received with enough
	 * bit errors that the current level of FEC still could not
	 * fix all of the problems - ie; after FEC, the checksum still
	 * failed.
	 */
	if( ch->error() ) {
		Packet::free(pktRx);
		set_nav(usec(eifs));
		goto done;
	}

	/*
	 * IEEE 802.11 specs, section 9.2.5.6
	 *	- update the NAV (Network Allocation Vector)
	 */
	if(dst != index) {
		set_nav(mh->dh_duration);
	}

        /* tap out */
        if (tap && type == MAC_Type_Data &&
            MAC_Subtype_Data == subtype) tap->tap(pktRx);

	/*
	 * Address Filtering
	 */
	if(dst != index && dst != MAC_BROADCAST) {
		/*
		 *  We don't want to log this event, so we just free
		 *  the packet instead of calling the drop routine.
		 *
		 *  I hate it when the comments don't match the code.
		 *  I'm not sure what's supposed to be happening here. -dam 4/12/99
		 */
		discard(pktRx, "---");
		goto done;
	}

	switch(type) {

	case MAC_Type_Management:
		discard(pktRx, DROP_MAC_PACKET_ERROR);
		goto done;
		break;

	case MAC_Type_Control:

		switch(subtype) {

		case MAC_Subtype_RTS:
			recvRTS(pktRx);
			break;

		case MAC_Subtype_CTS:
			recvCTS(pktRx);
			break;

		case MAC_Subtype_ACK:
			recvACK(pktRx);
			break;

		default:
			fprintf(stderr, "Invalid MAC Control Subtype %x\n",
				subtype);
			exit(1);
		}
		break;

	case MAC_Type_Data:

		switch(subtype) {

		case MAC_Subtype_Data:
			recvDATA(pktRx);
			break;

		default:
			fprintf(stderr, "Invalid MAC Data Subtype %x\n",
				subtype);
			exit(1);
		}
		break;

	default:
		fprintf(stderr, "Invalid MAC Type %x\n", subtype);
		exit(1);
	}

	done:

	pktRx = 0;
	rx_resume();
}


void
Mac802_11::recvRTS(Packet *p)
{
	struct rts_frame *rf = (struct rts_frame*)p->access(off_mac_);

	if(tx_state != MAC_IDLE) {
		discard(p, DROP_MAC_BUSY);
		return;
	}

	/*
	 *  If I'm responding to someone else, discard this RTS.
	 */
	if(pktCTRL) {
		discard(p, DROP_MAC_BUSY);
		return;
	}

	sendCTS(ETHER_ADDR(rf->rf_ta), rf->rf_duration);

	/*
	 *  Stop deferring - will be reset in tx_resume().
	 */
	if(mhDefer.busy()) mhDefer.stop();

	tx_resume();

	mac_log(p);
}


void
Mac802_11::recvCTS(Packet *p)
{
	if(tx_state != MAC_RTS) {
		discard(p, DROP_MAC_INVALID_STATE);
		return;
	}

	assert(pktRTS);
	Packet::free(pktRTS); pktRTS = 0;

	assert(pktTx);

	mhSend.stop();

	/*
	 * The successful reception of this CTS packet implies
	 * that our RTS was successful.  Hence, we can reset
	 * the Short Retry Count and the CW.
	 */
	ssrc = 0;
	rst_cw();

	tx_resume();

	mac_log(p);
}

void
Mac802_11::recvDATA(Packet *p)
{
	struct hdr_mac802_11 *dh = HDR_MAC11(p);
	u_int32_t dst, src, size;

	{	struct hdr_cmn *ch = HDR_CMN(p);

		dst = ETHER_ADDR(dh->dh_da);
		src = ETHER_ADDR(dh->dh_sa);
		size = ch->size();

		/*
		 * Adjust the MAC packet size - ie; strip
		 * off the mac header
		 */
		ch->size() -= ETHER_HDR_LEN11;
		ch->num_forwards() += 1;
	}

	/*
	 *  If we sent a CTS, clean up...
	 */
	if(dst != MAC_BROADCAST) {
		if(size >= macmib->RTSThreshold) {
			if (tx_state == MAC_CTS) {
				assert(pktCTRL);
				Packet::free(pktCTRL); pktCTRL = 0;

				mhSend.stop();

				/*
				 * Our CTS got through.
				 */
				ssrc = 0;
				rst_cw();
			}
			else {
				discard(p, DROP_MAC_BUSY);
				return;
			}
			sendACK(src);
			tx_resume();
		}
		/*
		 *  We did not send a CTS and there's no
		 *  room to buffer an ACK.
		 */
		else {
			if(pktCTRL) {
				discard(p, DROP_MAC_BUSY);
				return;
			}
			sendACK(src);
			if(mhSend.busy() == 0)
				tx_resume();
		}
	}

	/* ============================================================
	    Make/update an entry in our sequence number cache.
	   ============================================================ */
	if(dst != MAC_BROADCAST) {
		Host *h = &cache[src];

		if(h->seqno && h->seqno == dh->dh_scontrol) {
			discard(p, DROP_MAC_DUPLICATE);
			return;
		}
		h->seqno = dh->dh_scontrol;
	}

	/*
	 *  Pass the packet up to the link-layer.
	 *  XXX - we could schedule an event to account
	 *  for this processing delay.
	 */
	p->incoming = 1;
	recvtarget_->recv(p, (Handler*) 0);
}


void
Mac802_11::recvACK(Packet *p)
{
	struct hdr_cmn *ch = HDR_CMN(p);

	if(tx_state != MAC_SEND) {
		discard(p, DROP_MAC_INVALID_STATE);
		return;
	}

	assert(pktTx);
	Packet::free(pktTx); pktTx = 0;

	mhSend.stop();

	/*
	 * The successful reception of this ACK packet implies
	 * that our DATA transmission was successful.  Hence,
	 * we can reset the Short/Long Retry Count and the CW.
	 */
	if((u_int32_t) ch->size() <= macmib->RTSThreshold)
		ssrc = 0;
	else
		slrc = 0;

      /*
       * Backoff before sending again.
       */
	rst_cw();
	assert(mhBackoff.busy() == 0);
	mhBackoff.start(cw, is_idle());

	tx_resume();

	mac_log(p);
}

